
#include "util/graphs.hpp"
#include "util/seq.hpp"
#include "solv/options.hpp"
#include "solv/caterpillar.hpp"
#include "math.h"
#include <cstdlib> // for rand()
#include <unistd.h> // for getpid

void usage(const char* progname, std::ostream& o){
  o << "usage: " << progname << " ftree <file to read> [more opts]\t- read tree from file" << std::endl;
  o << "       " << progname << " scat <file to read> [more opts]\t- read sequence from file" << std::endl;
  o << "       " << progname << " rtree <#vertices> [more opts]\t- create random tree"<< std::endl;
  o << "       " << progname << " rcat <#vertices> [more opts]\t- create random caterpillar"<< std::endl;
  exit(1);
}

status::tree* get_random_tree(const size_t num_vertices){
  status::vertex* vertex_nr[num_vertices];
  status::vertex* root = new status::vertex();
  status::tree* t = new status::tree(root);

  srand(time(NULL) * getpid());

  vertex_nr[0] = root;
  for(uint i = 1; i < num_vertices; ++i)
    vertex_nr[i] = t->add_vertex(vertex_nr[rand() % i]);
  
  return t;
}


#define percentage_backbone 30
status::tree* get_random_caterpillar(const size_t num_vertices){
  status::vertex* vertex_nr[num_vertices];
  status::vertex* root = new status::vertex();
  status::tree* t = new status::tree(root);

  srand(time(NULL) * getpid());

  // create backbone
  size_t num_backbone = uint(num_vertices*percentage_backbone/100) + (rand() % uint(num_vertices*(100-percentage_backbone)/100));
  std::cout << "selected "<<num_backbone<<" vertices for backbone" << std::endl;
  vertex_nr[0] = root;
  for(uint i = 1; i < num_backbone; ++i)
    vertex_nr[i] = t->add_vertex(vertex_nr[i-1]);

  // create leaves
  for(uint i = num_backbone; i < num_vertices; ++i)
    vertex_nr[i] = t->add_vertex(vertex_nr[rand() % num_backbone]);
  
  return t;
}


const std::pair<string, int> _requires_params[] = {
  { "ftree", 1 },
  { "scat", 1 },
  { "rtree",  1 },
  { "rcat",  1 },
};
// global arguments with their parameters
std::map<string, std::vector<string> > arguments;

void parse_args(int argc, char** argv, status::sequence_t& I, status::solv_options& opts){
  int arg_ptr;
  std::map<std::string, int>  requires_params(std::begin(_requires_params), std::end(_requires_params));
  arg_ptr = 1;
  while(arg_ptr < argc){
    const std::string arg(argv[arg_ptr++]);
    // if the argument is not registered in requires_args, then exit with usage
    if(requires_params.find(arg) == requires_params.end()) usage(argv[0], std::cerr);
    // if there are not enough parameters for this argument
    if(argc < arg_ptr + requires_params[arg]) usage(argv[0], std::cerr);
    // otherwise fill the argument map
    std::vector<string> params(requires_params[arg]);
    for(int i = 0; i < requires_params[arg]; ++i)
      params[i] = argv[arg_ptr++];
    arguments.insert(make_pair(arg, params));
  }
}

int main(int argc, char** argv)
{
  status::solv_options opts;
  status::sequence_t I;
  status::tree* t;
  bool is_caterpillar = false;

  // parse the arguments, filling 'arguments'
  parse_args(argc, argv, I, opts);

  if(arguments.find("rtree") != arguments.end()){
    // create a random tree
    t = get_random_tree(atoi(arguments["rtree"][0].c_str()));
  } else if(arguments.find("rcat") != arguments.end()) {
    // create a random caterpillar
    t = get_random_caterpillar(atoi(arguments["rcat"][0].c_str()));
    is_caterpillar = true;
  } else if(arguments.find("ftree") != arguments.end()) {
    // create a graph from file
    t = new status::tree();
    t->read_from_file(arguments["ftree"][0]);
    is_caterpillar = detect_caterpillar(*t);
  } else if(arguments.find("scat") != arguments.end()) {
    // create a graph from file
    status::sequence_t s(status::read_sequence_from_file(arguments["scat"][0]));
    std::cout << "status sequence: " << s << std::endl;

    status::tree* result = status::stati_to_caterpillar(s);
    if(result){
      std::cout << "reconstructed:" << std::endl << *result << std::endl << "largest list: "<<status::get_set_list_max()<<endl;
      exit(0);
    } else {
      std::cout << "could not reconstruct the graph" << std::endl;
      exit(1);
    }
  } else usage(argv[0], std::cerr);

  cout << "writing tree to .tmp for reference" << endl;
  t->write_to_file(".tmp");

  std::cout << *t << endl;
  cout << "computing stati"<< endl;
  const status::sequence_t s(status::compute_stati(*t));
  // reroot t at its median
  cout << "computing median"<<endl;
  status::vertex* median = status::compute_median(*t);
  t->clear_data();
  cout << "rerooting tree at median" <<endl;
  t->reroot(median);

  std::cout << *t << endl;
  std::cout << "status sequence: " << s << std::endl;

  status::tree* result;
  if(is_caterpillar){
    result = status::stati_to_caterpillar(s);
  } else {
    std::cout << "this is not a caterpillar..."<<std::endl;
    result = NULL;
  }
  if(result){
    std::cout << "reconstructed:" << std::endl << *result << std::endl << "largest list: "<<status::get_set_list_max()<<std::endl;
    status::sequence_t check(status::compute_stati(*result));
    std::cout << "recheck stati "<<check<<": "<<(status::equal(s, check) ? "match! Good job :)" : "!!! NO MATCH !!!")<<std::endl;
  } else {
    std::cout << "could not reconstruct the graph" << std::endl;
  }
}
